def get_inverse_char(num,string):
    
    inverse_num = len(string)-num
    

    return -(inverse_num+1)


from collections import defaultdict
import itertools
def longest_common_prefix(str1, str2):
    """Find the longest common prefix between two strings."""
    min_length = min(len(str1), len(str2))
    for i in range(min_length):
        if str1[i] != str2[i]:
            return str1[:i]
    return str1[:min_length]

def longest_common_suffix(str1, str2):
    """Find the longest common suffix between two strings."""
    str1 = str1[::-1]
    str2 = str2[::-1]
    return longest_common_prefix(str1, str2)[::-1]

def get_inverse_char(index, str):
    """Return the character from the end based on the index from the start."""
    return len(str) - index - 1
def longest_common_prefix(str1, str2):
    """Find the longest common prefix between two strings."""
    min_length = min(len(str1), len(str2))
    for i in range(min_length):
        if str1[i] != str2[i]:
            return str1[:i]
    return str1[:min_length]

def longest_common_suffix(str1, str2):
    """Find the longest common suffix by reversing strings and using longest_common_prefix."""
    str1 = str1[::-1]
    str2 = str2[::-1]
    return longest_common_prefix(str1, str2)[::-1]

def get_inverse_char(index, str):
    """Return the character from the end based on the index from the start."""
    return len(str) - index - 1

def check_inverse_char_presence(header, comp_header):
    """Check the presence of each inverse character from 'header' in 'comp_header'."""
    inverse_presence = []
    for i in range(len(header)):
        inverse_char = header[get_inverse_char(i, header)]
        if inverse_char in comp_header:
            inverse_presence.append((inverse_char, True))
        else:
            inverse_presence.append((inverse_char, False))
    return inverse_presence
def find_unique_matches(list1, list2):
    """
    Finds unique substring matches between two lists of strings.
    
    Parameters:
        list1 (list): The first list of strings.
        list2 (list): The second list of strings.
        
    Returns:
        dict: A dictionary of unique matches where each key-value pair is a unique relationship.
    """
    matches = {}

    # Check each string in the first list against each string in the second list
    for str1 in list1:
        for str2 in list2:
            if str1 in str2 or str2 in str1:
                # Check if either string already has a match registered
                if str1 not in matches and str2 not in matches.values():
                    matches[str1] = str2
                elif str1 in matches:
                    # If already matched but found another, remove and note as non-unique
                    if matches[str1] != str2:
                        del matches[str1]
                        break
                elif str2 in matches.values():
                    # Find the key for the current value str2 and remove it for being non-unique
                    key_to_remove = [key for key, value in matches.items() if value == str2][0]
                    del matches[key_to_remove]

    return matches


def calculate_match_score(str1, str2):
    """
    Calculates a match score based on the length of the longest common substring
    relative to the string lengths.
    """
    if str1 and str2:#[float(0),0,None,'nan','NaN']
        if str1 in str2:
            return len(str1) / len(str2)
        elif str2 in str1:
            return len(str2) / len(str1)
    return 0

def find_best_unique_matches(list1, list2):
    """
    Identifies the best matches based on substring content and uniqueness.
    
    Parameters:
        list1 (list): The first list of strings.
        list2 (list): The second list of strings.
        
    Returns:
        dict: A dictionary with best unique matches.
    """
    potential_matches = {}
    for str1 in list1:
        for str2 in list2:
            score = calculate_match_score(str1, str2)
            if score > potential_matches.get((str1, str2), 0):
                potential_matches[(str1, str2)] = score

    # Filter to ensure uniqueness and best scores
    best_matches = {}
    for (str1, str2), score in potential_matches.items():
        if str1 not in best_matches and all(str2 != match for match in best_matches.values()):
            best_matches[str1] = str2
        elif str1 in best_matches:
            current_match = best_matches[str1]
            if potential_matches[(str1, current_match)] < score:
                best_matches[str1] = str2

    return best_matches
def find_common_substrings(strings, min_length=1):
    """
    Finds common substrings among a list of strings.
    
    Parameters:
        strings (list): List of input strings.
        min_length (int): Minimum length of substring to consider.
        
    Returns:
        set: A set containing all substrings found in more than one string.
    """
    common_substrings = defaultdict(int)
    substring_set = set()

    # Generate all possible substrings for each string
    for string in strings:
        for start in range(len(string)):
            for end in range(start + min_length, len(string) + 1):
                substring = string[start:end]
                common_substrings[substring] += 1
                if common_substrings[substring] > 1:
                    substring_set.add(substring)

    return substring_set

def remove_common_substrings(strings, common_subs):
    """
    Removes common substrings from a list of strings.
    
    Parameters:
        strings (list): List of strings to process.
        common_subs (set): Set of substrings to remove.
        
    Returns:
        list: Updated list of strings with common substrings removed.
    """
    updated_strings = []
    for string in strings:
        for sub in common_subs:
            string = string.replace(sub, '')
        updated_strings.append(string)
    return updated_strings

comp_headers = """Source.Name	INPUT: Extra 3	Date Pulled	Code	INPUT: Extra 1	INPUT: First Name	INPUT: Middle Name	INPUT: Last Name	INPUT: Address 1	INPUT: Address 2	INPUT: City	INPUT: State	ZIP	INPUT: Extra 4	INPUT: Extra 5	INPUT: Extra 6	PH: Phone1	EMAIL: Email1""".split('\t')
df = "/home/gamook/Documents/pythonScripts/test_grounds/all_excels/originals/ALL Time Data for Solar.xlsx"
df = get_df(df)
headers = df.columns.tolist()
headlower = [str(head).lower() for head in headers]
comp_lowers = [str(comp).lower() for comp in comp_headers]
# Lowercase versions of headers

# Find and print the best unique matches
best_matches = {}
matches = main(comp_lowers, headlower)
for key,value in matches.items():
    if key in comp_lowers:
        comp_lowers.remove(key)
    if value in comp_lowers:
        comp_lowers.remove(value)
        
    if key in headlower:
        headlower.remove(key)
    if value in headlower:
        headlower.remove(value)
    best_matches[key]=value
input(best_matches)
matches = find_unique_matches(comp_lowers, headlower)

input(matches)
for key,value in matches.items():
    if key in comp_lowers:
        comp_lowers.remove(key)
    if key in headlower:
        headlower.remove(key)
    if value in comp_lowers:
        comp_lowers.remove(value)
    if value in headlower:
        headlower.remove(value)
    best_matches[key]=value

input(comp_lowers)
results = []
highest = [None,None]
for i,comp_lower in enumerate(comp_lowers):
    if i!= 0:
        prefix = longest_common_prefix(comp_lowers[i-1], comp_lowers[i])
        common = [obj for obj in comp_lowers if obj.startswith(prefix)]
        if highest[0] == None:
            highest=[prefix,common]
        else:
            if len(prefix)>len(highest[0]):
                highest=[prefix,common]
comp_lowers = [obj[len(prefix):] for obj in highest[1]]+[obj for obj in comp_lowers if obj not in highest[1]]
comp_lowers_pre = highest[1]+[obj for obj in comp_lowers if obj not in highest[1]]
highest = [None,None]
for i,headlow in enumerate(headlower):
    if i!= 0:
        prefix = longest_common_prefix(headlower[i-1], headlower[i])
        common = [obj for obj in headlower if obj.startswith(prefix)]
        if highest[0] == None:
            highest=[prefix,common]
        else:
            if len(prefix)>len(highest[0]):
                highest=[prefix,common]
headlower = [obj[len(prefix):] for obj in highest[1]]+[obj for obj in headlower if obj not in highest[1]]

headlowers_pre = highest[1]+[obj for obj in headlower if obj not in highest[1]]
matches = main(comp_lowers, headlower)
for key,value in matches.items():
    if key in comp_lowers_pre:
        comp_lowers_pre.remove(key)
    if key in headlowers_pre:
        headlowers_pre.remove(key)
    if value in comp_lowers_pre:
        comp_lowers_pre.remove(value)
    if value in headlowers_pre:
        headlowers_pre.remove(value)
    best_matches[key]=value
input(matches)
results = []
# Comparing each string in the computed headers list against each string in the original headers list
input(get_closest_headers(df,comp_lowers))
for comp_header in comp_lowers:
    for header in headlower:
        prefix = longest_common_prefix(comp_header, header)
        suffix = longest_common_suffix(comp_header, header)

        # Calculate non-matching middle part indices
        middle_start = len(prefix)
        middle_end = len(comp_header) - len(suffix)
        middle = comp_header[middle_start:middle_end] if middle_end > middle_start else ''

        # Check for inverse character presence
        inverse_presence = check_inverse_char_presence(header, comp_header)

        results.append({
            'header': header,
            'comp_header': comp_header,
            'common_prefix': calculate_match_score(prefix, header),
            'common_suffix': calculate_match_score(suffix, header),
            'middle': calculate_match_score(middle, header),
            'inverse_presence':calculate_match_score([pres[0] for pres in inverse_presence if pres[1]], inverse_presence),
            'unique_match':find_unique_matches([header], comp_lowers)
        })

input(results)



import numpy as np
from abstract_pandas import get_df,get_closest_headers,append_unique_to_excels,get_address_header_association
from abstract_pandas.landManager import get_bp
from .word_compare import *
from abstract_utilities import get_any_value,get_closest_boudary
import re,requests, httpx, asyncio
from openlocationcode import openlocationcode as olc
from .functions import *
def get_default_google_api_key():
    return get_env_value(key="google_maps_api_key")
def get_closest_exact_boudary(D_lat, D_lon, boundary_polygon=None,api_key=None):
    api_key = api_key or get_default_google_api_key()
    boundary_polygon = get_bp(boundary_polygon)
    shortest,coordinates = get_closest_boudary(D_lat, D_lon, boundary_polygon)
    address_2 = get_address_from_latlngs(coordinates[0],coordinates[1])
    return get_lat_long(address_2,api_key)
def get_from_distance_response(json_response):
    text_finds = get_any_value(json_response, 'text')
    if text_finds:
        text_find_list = make_list(text_finds)
        if text_find_list and isinstance(text_find_list, list) and len(text_find_list) > 0:
            text_find_list = text_find_list[0]
        if text_find_list and isinstance(text_find_list, str):
            return text_find_list.split(' ')[0]
    return text_finds
def get_lat_long(address,boundary_polygon=None, api_key=None):
    api_key =api_key or get_default_google_api_key()
    boundary_polygon = get_bp(boundary_polygon)
    
    base_url = "https://maps.googleapis.com/maps/api/geocode/json"
    params = {
        "address": address,
        "key": api_key
    }
    
    response = requests.get(base_url, params=params)
    if response.status_code == 200:
        results = response.json()['results']
        if results:
            location = results[0]['geometry']['location']
            return location['lat'], location['lng']
        else:
            return None, None
    else:
        return None, None
def get_address_from_latlngs(lat, lng, api_key=None):
    api_key = api_key or get_default_google_api_key()
    # Endpoint URL
    base_url = "https://maps.googleapis.com/maps/api/geocode/json"
    
    # Parameters for the API request
    params = {
        "latlng": f"{lat},{lng}",
        "key": api_key
    }
    
    # Send request and get response data
    response = requests.get(base_url, params=params)
    
    # Convert response to JSON format

    results = response.json()
    
    # Check if any results were returned
    if results["status"] == "OK":
        # Return the formatted address of the first result
        address = results["results"][0]["formatted_address"]
        return f"{address.split(' ')[0].split('+')[0]} {' '.join(address.split(' ')[1:])}"
    else:
        # Return error message if no results found
        return "No address found for this latitude and longitude"
async def calculateDistance(origin, destination,travelMode = 'DRIVING',api_key=None,print_response = True,return_all = False):
    api_key =api_key or get_default_google_api_key()
    base_url = f"https://maps.googleapis.com/maps/api/distancematrix/json?origins={origin}&destinations={destination}&travelMode={travelMode}&units=imperial&key={api_key}"
    
    async with httpx.AsyncClient() as client:
        response = await client.get(base_url)
        
        if response.status_code != 200:
            # Handle error or retry as needed
            if print_response:
                print(response.text)
            return None
        if print_response:
            print(response.text)
        response = response.json()
        # Your existing logic to extract distance
        if return_all == False:
            response = get_from_distance_response(response)
        return response
async def get_distance_calculation(address, boundary_polygon=None,api_key=None):
    boundary_polygon = get_bp(boundary_polygon)
    if isinstance(address,str):
        dest_latitude,dest_longitude = get_lat_long(address,boundary_polygon)
    in_city_limit = is_point_inside_polygon(dest_latitude, dest_longitude, boundary_polygon)
    if in_city_limit:
        return 0
    coordinates = get_closest_exact_boudary(dest_latitude,dest_longitude,boundary_polygon)
    src_latitude,src_longitude= coordinates[0],coordinates[1]
    distance = await calculateDistance(f"{src_latitude} {src_longitude}",f"{dest_latitude} {dest_longitude}",travelMode="DRIVING")
    return distance

def get_address_header_association(headers_js,df,index,spec_headers=False):
    print(headers_js)
    if not isinstance(headers_js,dict):
        headers_js = get_closest_headers(headers_js)
    if isinstance(headers_js,dict):
        address = f"{df[headers_js['address']].loc[index]} {df[headers_js['city']].loc[index]} {df[headers_js['state']].loc[index]} {df[headers_js['zip']].loc[index]}"
        if spec_headers and isinstance(spec_headers,list):
            address=""
            for spec_header in spec_headers:
                spec_header = headers_js.get(spec_header)
                if spec_header:                    
                    address+=f"{df[spec_header].loc[index]} "
            address = eatAll(address,' ')
        return address
async def calculate_distances(new_data, headers_js, api_key=None):
    api_key =api_key or get_default_google_api_key()
    distance = new_data.get('distance')
    address = None
    # Check if distance is 'NaN', None, an empty string, or zero (as string or number)
    if distance is None or distance == "" or isinstance(distance, str) and not distance.isdigit():
        address = get_address_header_association(headers_js, new_data)
        distance = await get_distance_calculation(address,boundary_polygon)
    elif isinstance(distance, (int, float)):  # Check if it is a number (int or float)
        if distance == 0 or math.isnan(distance):  # Explicit check for 0 and NaN
            address = get_address_header_association(headers_js, new_data)
            distance = await get_distance_calculation(address,boundary_polygon)
    elif isinstance(distance, str) and distance.isdigit():
        if int(distance) == 0:  # Additional check if the string number is zero
            address = get_address_header_association(headers_js, new_data)
            distance = await get_distance_calculation(address,boundary_polygon)

    return distance  # Return the processed or retrieved distance
async def get_distance_for_excel_row(df,file_path=None,row_itter=0,distance=None,row_data=None,address=None,headers_js=None,search_column=None,boundary_polygon=None, search_value=None,distance_header='distance'):
    boundary_polygon = get_bp(boundary_polygon)
    df=get_df(df or file_path,make_excel=row_data)
    if row_data==None:
       row_data = get_row_as_list(df,index=row_itter) or {}
    if address == None:
        address = get_address_header_association(headers_js,df,row_itter)
    if distance == None:
        distance = await get_distance_calculation(address, boundary_polygon)
    row_data[distance_header] = distance or 0
    return append_unique_to_excels(df,file_path = file_path, new_data = row_data,search_column=search_column, search_value=search_value)

async def add_distance_calculatuions_to_excel_data(excel_file_path,boundary_polygon=None,append=False,new_excel_path="new_excel_file.xlsx"):
    boundary_polygon = get_bp(boundary_polygon)
    api_key = api_key or get_default_google_api_key()
    if not os.path.isfile(excel_file_path):
        return
    all_distance = safe_read_from_json("all_distances.json")
    headers_js = get_closest_headers(excel_file_path)
    main_df = get_df(excel_file_path)
    for index, row in main_df.iterrows():
            zip_now = str(row.get('ZIP'))
            all_dst = all_distance.get(zip_now)
            if not is_number(all_dst):
                
                all_distance[zip_now]=await get_distance_calculation(get_address_header_association(headers_js,main_df,index),boundary_polygon=boundary_polygon)
            if is_number(all_distance[zip_now]):
                row['distance'] = all_distance[zip_now]
            else:
                dist = main_df.at[index, 'distance']
                zip_now = str(row.get('ZIP'))
                all_dst = all_distance.get(zip_now)
                if is_number(all_dst):
                    row['distance'] = str(all_dst)
                
                    
                elif str(dist) != str(all_dst) or not is_number(all_dst) or pd.to_numeric(dist, errors='coerce') == 0 or dist == 'nan' or pd.isna(dist) or not is_number(dist):
                    address = get_address_header_association(headers_js,main_df,index)
                    row['distance']=await get_distance_calculation(address, boundary_polygon=boundary_polygon)
                    all_distance[zip_now] = row['distance']
                    safe_dump_to_json(data=all_distance,file_path=-'all_distances.js')
            main_df = append_unique_to_excels(main_df, new_data = row,search_column=headers_js['external'], search_value=row.get(headers_js['external']),print_it=False)
            
            safe_excel_save(main_df,new_excel_path)

def convert_to_str_lalng(obj):
    if isinstance(obj,str):
        obj = obj.replace(',',' ').split(' ')
        return f"{obj[0]} {obj[-1]}"
    for ty in [set,list,tuple]:
        if isinstance(obj,ty):
            return f"{obj[0]} {obj[-1]}"
    return obj
def decode_plus_code(plus_code):
    decoded = olc.decode(plus_code)
    return (decoded.latitudeCenter, decoded.longitudeCenter)







import asyncio,json,shutil
from .distance_cals import *
def get_abs_dir():
    return os.path.dirname(os.path.abspath(__file__))
def get_json_data(file_path):
    if not os.path.isfile(file_path):
        safe_dump_to_file(data={},file_path=file_path)
    return safe_read_from_json(file_path)
def get_all_distance_data(data_file_path=None):
    data_file_path = data_file_path or get_all_distance_file_path()
    return get_json_data(file_path=data_file_path)
def save_all_distances(data,data_file_path=None):
    safe_dump_to_file(data=data, file_path=data_file_path or get_all_distance_file_path())
def get_all_distance_file_path():
    return os.path.join(get_abs_dir(),"all_distances.json")
def update_excel_to_column_values(df,update_header,value_header,values_js):
    df = get_df(df)
    df[update_header] = df[value_header].apply(lambda x: values_js.get(str(x)))
    return df
def get_all_excels_dir():
    excel_dir = os.path.join(get_abs_dir(),'all_excels')
    os.makedirs(excel_dir,exist_ok=True)
    return excel_dir
def get_original_excels_dir():
    curr_original_dir = os.path.join(get_all_excels_dir(),'originals')
    os.makedirs(curr_original_dir,exist_ok=True)
    return curr_original_dir
def get_filtered_excels_dir():
    curr_filtered_dir = os.path.join(get_all_excels_dir(),'filtered')
    os.makedirs(curr_filtered_dir,exist_ok=True)
    return curr_filtered_dir
def get_original_ext_dir(ext):
    original_ext_dir = os.path.join(get_original_excels_dir(),eatAll(ext,['.']))
    os.makedirs(original_ext_dir,exist_ok=True)
    return original_ext_dir
def get_filtered_ext_dir(ext):
    filtered_ext_dir = os.path.join(get_filtered_excels_dir(),eatAll(ext,['.']))
    os.makedirs(filtered_ext_dir,exist_ok=True)
    return filtered_ext_dir
def get_original_filepath(baseName):
    directory = os.path.dirname(baseName)
    if os.path.isdir(directory):
        baseName= os.path.basename(baseName)
    ext=os.path.splitext(baseName)
    return os.path.join(get_original_ext_dir(ext[-1]),baseName)
def get_filtered_filepath(baseName):
    directory = os.path.dirname(baseName)
    if os.path.isdir(directory):
        baseName= os.path.basename(baseName)
    ext=os.path.splitext(baseName)
    return os.path.join(get_filtered_ext_dir(ext[-1]),baseName)
def get_file_associations(file_path):
    original_file_path = get_original_filepath(file_path)
    filtered_file_path = get_filtered_filepath(file_path)
    if file_path != original_file_path and os.path.isfile(file_path):
        shutil.copy(file_path,original_file_path)
    return original_file_path,filtered_file_path
def check_if_none_val(value):
    if value in [0,'null',None,'None','nan','NaN','0.0',0.0]:
        return True
    return False
async def update_distances(df,update_header,value_header,values_js,data_file_path=None,boundary_polygon=None,api_key=None):
    """Asynchronously update distances in the DataFrame."""
    df = get_df(df)
    boundary_polygon = get_bp(boundary_polygon)
    headers_js = get_closest_headers(df,value_header,update_header)
    values_js = values_js or get_all_distance_data(data_file_path=data_file_path)
    for index, row in df.iterrows():
        # Check if the current cell in update_header column is not a number
        
        if not is_number(row[update_header]) or str(row[update_header]) in ['nan','NaN','None',None,0,'0.0']:
            # Use the value in the value_header column as the key for the json_dict
            key = df.at[index, value_header]
            value = values_js.get(str(key))
            # Check if the key exists in json_dict and if its value can be converted to a number
            if value and is_number(value) and (float(value) != float(0) or str(row[update_header]) in ['nan','NaN','None',None,0,'0.0']):
                # Update the cell with the value from json_dict
                df.at[index, update_header] = value
            else:
                # Mark the cell for update
                if key != 'nan':
                    values_js[key] = await get_distance_calculation(get_address_header_association(headers_js, df, index),boundary_polygon,api_key)
                    if check_if_none_val(values_js[key]):
                        for i in range(1,3):
                            values_js[key] = await get_distance_calculation(get_address_header_association(headers_js, df, index,['city','state','zip'][-i:]),boundary_polygon,api_key)
                            if not check_if_none_val(values_js[key]):
                                break
                    df.at[index, update_header] = values_js[key]
                    save_all_distances(data=values_js,data_file_path=data_file_path)
    # Write the distances back to file once all updates are completed
    return df,values_js
async def update_excel(df,update_header,value_header,values_js,data_file_path=None):
    df = get_df(df)
    values_js = values_js or get_all_distance_data(data_file_path=data_file_path)
    df = add_or_update_headers(df,'distance')
    values_js = convert_values(values_js,[float,str],elim=True)
    values_js = convert_keys(values_js,[float,int,str])
    df = convert_column(df,'ZIP',[float,int,str])
    df = convert_column(df,'distance',[float,str])
    df,values_js = await update_distances(df,update_header,value_header,values_js,data_file_path=data_file_path)
    return df,values_js
def convert_column(df,column,ty):
    df = get_df(df)
    ty= make_list(ty)
    df = add_or_update_headers(df,column)
    for typ in ty:
        try:
            df[column] = df[column].astype(typ)
        except:
            pass
    return df
def convert_keys(values_js,ty):
    new_js = {}
    ty= make_list(ty)
    for key,values in values_js.items():
        if key not in [None,'None','nan','NaN']:
            for typ in ty:
                key = typ(key)
        new_js[key]=values
    return new_js
def convert_values(values_js,ty,elim=False):
    new_js = {}
    ty= make_list(ty)
    for key,values in values_js.items():
        if values not in [None,'None','nan','NaN']:
            for typ in ty:
                
                values = typ(values)
            new_js[key]=values
    return new_js
def convert_all(df,values_js,column,ty):
    ty= make_list(ty)
    for typ in ty:
        df = convert_column(df,column,typ)
    return df,values_js
def update_all_values_js(df,value_header,values_js={},data_file_path=None):
    data_file_path = data_file_path or get_all_distance_file_path()
    original_file_path,filtered_file_path = get_file_associations(df)
    df = get_df(df)
    values_js = values_js or get_all_distance_data(data_file_path=data_file_path)

    values_js = convert_values(values_js,[float,str],elim=True)
    values_js = convert_keys(values_js,[float,int,str])
    df = convert_column(df,'ZIP',[float,int,str])
    df = convert_column(df,'distance',[str])
    needs_update = ~df[value_header].apply(lambda x: is_number(values_js.get(str(x))))
    update_indices = df[needs_update].index
    for i in update_indices:
        values_js[df['ZIP'][i]] = df['distance'][i]
    safe_excel_save(df,filtered_file_path)
    save_all_distances(data=values_js,data_file_path=data_file_path)
    
async def tryoit(directory=None,all_distance_path=None,destination_directory=None,ext='xlsx'):
    save_paths = []
    all_distance_path = all_distance_path or get_all_distance_file_path()
    directory = directory or get_original_ext_dir(ext)
    if os.path.isdir(directory):
        directory = os.listdir(directory)
    else:
        directory = make_list(directory)
    for excel_file_path in directory:
        values_js = get_json_data(all_distance_path)
        if not os.path.isfile(excel_file_path):
            excel_file_path = os.path.join(directory,excel_file_path)
        original_file_path,filtered_file_path = get_file_associations(excel_file_path)
        df,values_js= await update_excel(original_file_path,'distance','ZIP',values_js=values_js,data_file_path=all_distance_path)
        safe_dump_to_file(data=values_js, file_path=all_distance_path)
        safe_excel_save(df,filtered_file_path)
        save_paths.append(filtered_file_path)
        if destination_directory:
            move_file = os.path.join(destination_directory,os.path.basename(filtered_file_path))
            shutil.move(filtered_file_path,move_file)
            save_paths[-1]=move_file
    return save_paths
def associate_distances(directory=None,all_distance_path=None,destination_directory=None,ext='xlsx'):
    return asyncio.run(tryoit(directory=directory,all_distance_path=all_distance_path,destination_directory=destination_directory,ext=ext))



from collections import defaultdict
import itertools,re
from .excel_module import get_df,pd
def compare_strings(s1, s2):
    """
    Compares two strings to check if they are the same length and identical in all characters
    except possibly within numeric characters. Returns the differing numbers if found.

    Parameters:
        s1 (str): The first string to compare.
        s2 (str): The second string to compare.

    Returns:
        tuple: (bool, list) where bool indicates if the strings match under the given conditions,
               and list contains the pairs of differing numbers if any.
    """
    
    # Check if the strings are the same length
    if len(s1) != len(s2):
        return (False, [])

    # Find all numbers in both strings
    numbers_s1 = re.findall(r'\d+', s1)
    numbers_s2 = re.findall(r'\d+', s2)

    # Replace digits in both strings with '0' to compare non-numeric parts
    transformed_s1 = re.sub(r'\d', '0', s1)
    transformed_s2 = re.sub(r'\d', '0', s2)

    if transformed_s1 != transformed_s2:
        return (False, [])

    # Comparing the numbers in the same order they appear
    differing_numbers = []
    for num1, num2 in zip(numbers_s1, numbers_s2):
        if num1 != num2:
            differing_numbers.append((num1, num2))

    # If the non-numeric parts are identical and the numeric parts have differences
    if differing_numbers:
        return (True, differing_numbers)
    else:
        # If there are no numeric differences
        return (True, [])

def get_most_original_from_ls(string, values_list,string_list):
    strings = [string,'']
    for string_2 in values_list:
        strings[1] = string_2
        bool_comp,string_comp = compare_strings(string, string_2)
        
        if bool_comp and string_comp and bool_comp not in string_list:
            lowest = None
            for i,each in enumerate(string_comp[0]):
                each=int(each)
                if lowest == None or lowest[1] > each:
                    lowest=[i,each]
            return strings[lowest[0]]
    return string
def calculate_match_score(str1, str2):
    """
    Calculates a match score based on the length of the longest common substring
    relative to the string lengths.
    """
    if str1 and str2:#[float(0),0,None,'nan','NaN']
        if str1 in str2:
            return len(str1) / len(str2)
        elif str2 in str1:
            return len(str2) / len(str1)
    return 0

def find_best_unique_matches(list1, list2):
    """
    Identifies the best matches based on substring content and uniqueness.
    
    Parameters:
        list1 (list): The first list of strings.
        list2 (list): The second list of strings.
        
    Returns:
        dict: A dictionary with best unique matches.
    """
    potential_matches = {}
    for str1 in list1:
        for str2 in list2:
            score = calculate_match_score(str1, str2)
            if score > potential_matches.get((str1, str2), 0):
                potential_matches[(str1, str2)] = score

    # Filter to ensure uniqueness and best scores
    best_matches = {}
    for (str1, str2), score in potential_matches.items():
        if str1 not in best_matches and all(str2 != match for match in best_matches.values()):
            best_matches[str1] = str2
        elif str1 in best_matches:
            current_match = best_matches[str1]
            if potential_matches[(str1, current_match)] < score:
                best_matches[str1] = str2

    return best_matches
def find_commons_dna(list1, list2,new_matches={}):
    # Step 1: Remove common substrings within each list to simplify strings
    common_subs_list1 = find_common_substrings(list1)
    common_subs_list2 = find_common_substrings(list2)
    cleaned_list1 = remove_common_substrings(list1, common_subs_list1)
    cleaned_list2 = remove_common_substrings(list2, common_subs_list2)
    # Step 2: Calculate best matches based on cleaned lists
    best_matches = find_best_unique_matches(cleaned_list1, cleaned_list2)

    for key,value in best_matches.items():
        if key or value:
            i =0
            for clean_1 in cleaned_list1:       
                if clean_1 == key:
                    key = list1[i]
                    break
                i +=1 
            i =0
            for clean_2 in cleaned_list2:
                if clean_2 == value:
                    value = list2[i]
                    break
                i +=1   
            new_matches[key]=value
    
    for key,value in new_matches.items():
        if key in list1:
            list1.remove(key)
        if value in list1:
            list1.remove(value)
            
        if key in list2:
            list2.remove(key)
        if value in list2:
            list2.remove(value)
        best_matches[key]=value
    return new_matches,list1,list2
def find_common_substrings(strings, min_length=1):
    """
    Finds common substrings among a list of strings.
    
    Parameters:
        strings (list): List of input strings.
        min_length (int): Minimum length of substring to consider.
        
    Returns:
        set: A set containing all substrings found in more than one string.
    """
    common_substrings = defaultdict(int)
    substring_set = set()

    # Generate all possible substrings for each string
    for string in strings:
        for start in range(len(string)):
            for end in range(start + min_length, len(string) + 1):
                substring = string[start:end]
                common_substrings[substring] += 1
                if common_substrings[substring] > 1:
                    substring_set.add(substring)

    return substring_set

def remove_common_substrings(strings, common_subs):
    """
    Removes common substrings from a list of strings.
    
    Parameters:
        strings (list): List of strings to process.
        common_subs (set): Set of substrings to remove.
        
    Returns:
        list: Updated list of strings with common substrings removed.
    """
    updated_strings = []
    for string in strings:
        for sub in common_subs:
            string = string.replace(sub, '')
        updated_strings.append(string)
    return updated_strings

def get_closest_headers(df,*args,**kwargs):
    df=get_df(df)
    convert_list = ["address","city","state","zip","external","phone","distance","fico","middle name","last name","full name","email"]
    for arg in args:
        if isinstance(arg,list):
            convert_list+=arg
        elif isinstance(arg,pd.DataFrame):
            headers = arg.columns.tolist()
            convert_list+=headers
        else:
            convert_list.append(arg)
    convert_list=list(set(convert_list))
    headers = df.columns.tolist()
    closest_js={str(key):safe_get([item for item in headers if (key.lower() in item.lower() or item.lower() == key.lower())],0) or "" for key in convert_list}
    headers = [head for head in headers if head not in list(closest_js.values())]
    convert_list = [head for head in convert_list if head not in list(closest_js.keys())]
    for key,value in find_best_unique_matches(convert_list, headers):
        closest_js[key]=value
    headers = [head for head in headers if head not in list(closest_js.values())]
    convert_list = [head for head in convert_list if head not in list(closest_js.keys())]
    for key,values in closest_js.items():
        for comp_key in headers:
            if comp_key.lower() == key:
                closest_js[key]=comp_key
                break
        if closest_js[key] == "":
            matches = {}
            for comp_key in headers:
                if comp_key not in list(closest_js.values()):
                    for char in key:
                        if char in comp_key.lower():
                            if comp_key not in matches:
                                matches[comp_key]=[]
                            if len(matches[comp_key])==0:
                                matches[comp_key].append('')
                            if matches[comp_key][-1]+char in comp_key.lower():
                                matches[comp_key][-1]+=char
                            else:
                                matches[comp_key].append(char)
            for header,values in matches.items():
                len_header = len(header)
                highest=[0,'']
                for val in values:
                    if len(val)>highest[0]:
                        highest=[len(val),val]
                matches[header]={'perc':highest[0]/len_header,'value':highest[1]}
            highest=[0,'']
            for header,val in matches.items():
                if val['perc']>highest[0]:
                    highest=[val['perc'],header]
            closest_js[key] = get_most_original_from_ls(highest[-1],headers, closest_js.values())
    return closest_js




from time import time
import setuptools
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
setuptools.setup(
    name='abstract_distances',
    version='0.0.0.5',
    author='putkoff',
    author_email='partners@abstractendeavors.com',
    description="",
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.11',
      ],
    install_requires=['abstract_utilities', 'pandas', 'openpyxl','abstract_security','requests'],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    # Add this line to include wheel format in your distribution
    setup_requires=['wheel'],
)
